;;; tdbsf-db-file-mode.el --- TDBSF Database File mode

;; Copyright (c) 2001, 2004 Florent Rougon

;; Maintainer: Florent Rougon <flo@via.ecp.fr>

;; This file is part of TDBSF.

;; TDBSF is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; TDBSF is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with TDBSF; see the file COPYING. If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(defface font-tdbsf-url-face
  '((t
     (:weight bold)))
  "Used for displaying URLs in tdbsf-db-file-mode.")

(require 'browse-url)
(defvar tdbsf-db-file-url-button-keymap
  (let ((map (make-sparse-keymap)))
    (define-key map [mouse-2] 'browse-url-at-mouse)
;;; I don't put this enabled by default because it feels a bit strange to type
;;; a URL in a text file, hit Enter to go to the next line and have no newline
;;; inserted but the browser launched instead.
;;;    (define-key map "\r" 'browse-url-at-point)
    map)
  "Keymap for URL buttons in tdbsf-db-file-mode.")

(defun tdbsf-compare-versions (a b)
  "Compare two version numbers.
Compare version numbers A and B. Each version number should be made
of decimal digits separated by dots (any number of them is
acceptable). The version numbers are compared from left to right.

Return:
  *  0  if the version numbers are equal
  *  1  if A > B
  * -1  if A < B

Examples:

  21.3.1 = 21.3.1
  21.3.1 > 21.3 > 21
  21.3.1 > 21.3.0 > 21.2.9999 > 20.76576.9879 > 20.67.8783443.98798.9879"

  (let ((la (split-string a "\\."))
        (lb (split-string b "\\."))
        ia ib
        (continue t)
        res)
    (while continue
      (setq continue nil)
      (if la
          (progn
            (setq ia (string-to-number (car la) 10))
            (if lb
                (progn
                  (setq ib (string-to-number (car lb) 10))
                  (if (> ia ib)
                      (setq res 1)
                    (if (< ia ib)
                        (setq res -1)
                      (setq continue t
                            la (cdr la)
                            lb (cdr lb)))))
              (setq res 1)))
        (if lb
            (setq res -1)
          (setq res 0))))
    res))
              
(defvar tdbsf-db-file-font-lock-keywords
  (list (cons (if (and (locate-library "gnus-art")
                       (progn
                         (require 'gnus-art)
                         (boundp 'gnus-button-url-regexp)))
                  gnus-button-url-regexp
                ;; If we couldn't find gnus-button-url-regexp in gnus-art, use
                ;; its default value (well, the one when POSIX character
                ;; classes such as [:word:] are supported) in Gnus v5.10.6
                "\\b\\(\\(www\\.\\|\\(s?https?\\|ftp\\|file\\|gopher\\|nntp\\|news\\|telnet\\|wais\\|mailto\\|info\\):\\)\\(//[-a-z0-9_.]+:[0-9]*\\)?[-a-z0-9_=!?#$@~%&*+\\/:;.,[:word:]]+[-a-z0-9_=#$@~%&*+\\/[:word:]]\\)")
              `(0
                ;; With Emacs >= 21.3.50, it is possible to use constructs
                ;; such as '(face font-tdbsf-url-face mouse-face highlight) in
                ;; conjunction with font-lock-extra-managed-props. It is
                ;; already documented in Emacs 21.2 and 21.3 but unfortunately
                ;; doesn't work in these versions.
                ,(if (= (tdbsf-compare-versions emacs-version "21.3.50") -1)
                     '(progn
                        (add-text-properties
                         (match-beginning 0) (match-end 0)
                         `(mouse-face highlight
                                      keymap
                                      ,tdbsf-db-file-url-button-keymap))
                        ;; Now, return the face to use
                        'font-tdbsf-url-face)
                   `'(face font-tdbsf-url-face mouse-face highlight
                           keymap ,tdbsf-db-file-url-button-keymap))
                prepend nil)
              ))
  "Default expressions to highlight in tdbsf-db-file-mode.
The format of this variable is the same as described in the documentation of
the variable `font-lock-keywords'.")

(when (= (tdbsf-compare-versions emacs-version "21.3.50") -1)
  ;; Taken from font-lock.el.
  (eval-when-compile
    (defmacro tdbsf-save-buffer-state (varlist &rest body)
      "Bind variables according to VARLIST and eval BODY restoring buffer \
state."
      `(let* ,(append varlist
                      '((modified (buffer-modified-p)) (buffer-undo-list t)
                        (inhibit-read-only t) (inhibit-point-motion-hooks t)
                        (inhibit-modification-hooks t)
                        deactivate-mark buffer-file-name buffer-file-truename))
         ,@body
         (when (and (not modified) (buffer-modified-p))
           (set-buffer-modified-p nil)))))

  ;; Very much inspired from Emacs 21.2's font-lock-default-unfontify-region
  (defun tdbsf-db-file-font-lock-unfontify-region (beg end)
    (tdbsf-save-buffer-state
     nil
     (remove-text-properties beg end
                             (if font-lock-syntactic-keywords
                                 '(face nil
                                        syntax-table nil
                                        font-lock-multiline nil
                                        mouse-face nil
                                        keymap nil)
                               '(face nil
                                      font-lock-multiline nil
                                      mouse-face nil
                                      keymap nil))))))


(define-derived-mode tdbsf-db-file-mode text-mode "TDBSF Database File"
  "Major mode for displaying and editing of TDBSF db files.

This mode is derived from Text mode.

It is automatically entered in the appropriate buffer when you select a
record match in the *TDBSF Results Summary* buffer.
\\{tdbsf-db-file-mode-map}"
  (mapcar 'make-local-variable
          '(;; Contains, for each record in the buffer (not only matching
            ;; records), markers positioned at the beginning and end of the
            ;; "boh" and "eoh" parts of the header (i.e., 4 markers per
            ;; record)
            tdbsf-this-file-s-records-info
            ;; Contains, for each matching record, a marker positioned at the
            ;; beginning of the "boh" part of the header of that record, and
            ;; the Global Record Index of the record.
            tdbsf-this-file-s-matching-records-info
            tdbsf-matches-overlays
            tdbsf-session-id
            font-lock-defaults))
  (setq font-lock-defaults
        '(tdbsf-db-file-font-lock-keywords
          nil
          nil
          nil
          tdbsf-db-file-beginning-of-syntax
          (font-lock-mark-block-function . mark-paragraph)
          ))
  (setq font-lock-defaults
        (append font-lock-defaults
                ;; See the comments in tdbsf-db-file-font-lock-keywords
                (list
                 (if (= (tdbsf-compare-versions emacs-version "21.3.50") -1)
                     '(font-lock-unfontify-region-function
                       . tdbsf-db-file-font-lock-unfontify-region)
                   '(font-lock-extra-managed-props . (mouse-face keymap)))))))

(defun tdbsf-create-this-file-s-records-info-and-markers
  (matching-records-info-array buf gri)
  "Register information about the records in buffer BUF.
Create several variables (which should be local to BUF, if the caller
made its job properly) about the file that BUF is visiting (BUF
should be visiting a TDBSF database file):
  - tdbsf-session-id, containing the TDBSF session ID taken from the
    buffer named `tdbsf-summary-buffer-name';
  - tdbsf-matches-overlays to store future overlays that will be used
    to fontify matches in BUF;
  - tdbsf-this-file-s-records-info to store information about all
    records in BUF: it is an array of arrays (one for each record),
    where the contained arrays are of the form:

      [(b-boh . e-boh) (b-eoh . e-eoh)]

    and b-boh is a marker placed at the beginning of the \"boh\" part
    of the header of the record this arrays refers to, e-boh a marker
    placed at the end of the \"boh\" part of this header, b-eoh a
    marker placed at the beginning of its \"eoh\" part and e-eoh a
    marker placed at the end of its \"eoh\" part.
  - tdbsf-this-file-s-matching-records-info to store an array with an
    entry for each record of BUF that matched the search expression.
    Each element of this array is itself an array whose first element
    is a marker placed at the beginning of the \"boh\" part of the
    corresponding record's header and second element the Global
    Record Index of that record.

The value of GRI should be the Global Record Index of the first
matching record for the file that BUF is visiting."
  (let ((i -1)
        ;; `newgri' is simply better suited than `gri' for the algorithm
        (newgri (1- gri))
        (loop-count (length matching-records-info-array))
        (info-about-all-records-in-the-current-buffer
         (aref (cdr
                ;; The following expression contains all the information
                ;; associated to the file that buf is visiting (this is
                ;; "file-elt" in tdbsf.el:tdbsf-process-output).
                (aref (aref
                       (with-current-buffer
                           tdbsf-summary-buffer-name tdbsf-matching-records)
                       gri) 0)
                ) 0))
        marker
        )
    (with-current-buffer buf
      ;; These bindings are intended to be local to the buffer buf
      ;; (responsability of the caller)
      (setq tdbsf-session-id
            (with-current-buffer tdbsf-summary-buffer-name tdbsf-session-id)
            tdbsf-matches-overlays ()

            ;; Create markers placed at the beginning and end of the boh and
            ;; eoh parts of all record headers in buf and store them in an
            ;; array (no Global Record Index here since not all of these
            ;; records matched the search expression).
            tdbsf-this-file-s-records-info
            (vconcat
             (mapcar #'(lambda (record-info)
                         (vconcat (mapcar
                                   #'(lambda (index)
                                       (cons
                                        ;; beginning of the "boh" (if index = 0)
                                        ;; or "eoh" (if index = 1) part of the
                                        ;; header
                                        (prog1
                                            (setq
                                             marker
                                             (set-marker
                                              (make-marker)
                                              (car (aref record-info index))))
                                          (set-marker-insertion-type marker t))
                                        ;; end of the "boh" (if index = 0)
                                        ;; or "eoh" (if index = 1) part of the
                                        ;; header
                                        (prog1
                                            (setq
                                             marker
                                             (set-marker
                                              (make-marker)
                                              (cdr (aref record-info index))))
                                          (set-marker-insertion-type marker
                                                                     nil))
                                        ))
                                   '(0 1))))
                     info-about-all-records-in-the-current-buffer))

            tdbsf-this-file-s-matching-records-info
            (make-vector (length matching-records-info-array) nil))

      ;; Create markers placed at the beginning of matching records in buf and
      ;; store them in an array (`tdbsf-this-file-s-matching-records-info')
      ;; along with the corresponding Global Record Index.
      (while (< (setq i (1+ i)) loop-count)
        (aset tdbsf-this-file-s-matching-records-info i
              (vector
               (prog1
                   (setq marker
                         (set-marker
                          (make-marker)
                          ;; We set the marker at the beginning of the "boh"
                          ;; part of the header
                          (aref (car (aref matching-records-info-array i)) 0)))
                 (set-marker-insertion-type marker t))
               (setq newgri (1+ newgri))
               )))
      )))


;;;; Functions to determine the start or end of a record identified by a
;;;; buffer position in that record; functions to move to the start or end of
;;;; the "current" record (the one "containing" the point).

(defun tdbsf-start-of-record (position)
  "Return the buffer position of the start of a record.
Return the position of the start of the last (i.e., highest numbered)
record that starts before, or at, the buffer position POSITION.

If there is no such record, return nil.

This function must be called in a buffer where the local variable
`tdbsf-this-file-s-records-info' has been initialized."
  (let ((i (length tdbsf-this-file-s-records-info))
        (found nil)
        res)
    (while (and (not found) (>= (setq i (1- i)) 0))
      (if (<=
           (marker-position
            (car (aref (aref tdbsf-this-file-s-records-info i) 0)))
           position)
          (setq found t
                res (marker-position
                     (car (aref
                           (aref
                            tdbsf-this-file-s-records-info i) 0))))))
    (if found
        res
      nil)))

(defun tdbsf-end-of-record (position)
  "Return the buffer position of the end of a record.
Return the position of the start of the first (i.e., lowest numbered)
record that starts after the buffer position POSITION.

If there is no such record, return nil.

This function must be called in a buffer where the local variable
`tdbsf-this-file-s-records-info' has been initialized."
  (let ((i -1)
        (found nil)
        res)
    (while (and (not found) (< (setq i (1+ i))
                               (length tdbsf-this-file-s-records-info)))
      (if (>
           (marker-position
            (car (aref (aref tdbsf-this-file-s-records-info i) 0)))
           position)
          (setq found t
                res (marker-position
                     (car (aref
                           (aref
                            tdbsf-this-file-s-records-info i) 0))))))
    (if found
        res
      nil)))

(defun tdbsf-goto-start-of-record ()
  "Move point to beginning of current record.
Move the point to the beginning of the last (i.e., highest numbered) record
that starts before or at point."
  (interactive)
  (let ((start-of-record (tdbsf-start-of-record (point))))
    (if start-of-record
        (goto-char start-of-record)
      ;; This will happen if the user inserts some text before the first
      ;; record and calls this function with the point before the start of
      ;; what was the first record (the inserted text could be a valid new
      ;; record, but we don't know about it; we only know about the records
      ;; the TDBSF engine made us aware of; this is to avoid having two
      ;; supposedly equivalent TDBSF file parsers, one in the TDBSF engine and
      ;; one in this Emacs interface.
      (error "Unable to find a record that starts before %d" (point)))))

(defun tdbsf-db-file-beginning-of-syntax ()
  (let ((start-of-record (if (boundp 'tdbsf-this-file-s-records-info)
                             (tdbsf-start-of-record (point))
                           nil)))
    (if start-of-record
        (goto-char start-of-record)
      (goto-char (point-min)))))

(defun tdbsf-goto-end-of-record ()
  "Move point to end of current record.
Move the point to the end of the first (i.e., lowest numbered) record
that starts after point."
  (interactive)
  (let ((end-of-record (tdbsf-end-of-record (point))))
    (if end-of-record
        (goto-char end-of-record)
      (goto-char (point-max)))))


(defun tdbsf-fontify-matches (matching-records-info-array buf)
  (require 'tdbsf-db-file-faces)
  (let* ((and-faces (make-vector tdbsf-db-file-and-faces-count nil))
         (near-faces (make-vector tdbsf-db-file-near-faces-count nil))
         face-names-array
         face-name-index
         and-face-index
         near-face-index
         (i -1)
         overlay)
    ;; Generate face names arrays
    (while (< (setq i (1+ i)) tdbsf-db-file-and-faces-count)
      (aset and-faces i (intern (concat "tdbsf-db-file-and-face-"
                                        (number-to-string (1+ i))))))
    (setq i -1)
    (while (< (setq i (1+ i)) tdbsf-db-file-near-faces-count)
      (aset near-faces i (intern (concat "tdbsf-db-file-near-face-"
                                        (number-to-string (1+ i))))))
    (with-current-buffer buf
      ;; Generate the overlays and give them the desired properties
      (mapcar
       #'(lambda (record-info)
           ;; Reset the face numbers for each record
           (setq and-face-index -1
                 near-face-index -1)
           (mapcar
            #'(lambda (group)
                (if (string-equal "&" (car group))
                    (setq and-face-index (mod (1+ and-face-index)
                                              tdbsf-db-file-and-faces-count)
                          face-names-array and-faces
                          face-name-index and-face-index)
                  (setq near-face-index (mod (1+ near-face-index)
                                             tdbsf-db-file-near-faces-count)
                        face-names-array near-faces
                        face-name-index near-face-index))
                (mapcar
                 #'(lambda (match-span)
                     (overlay-put
                      (setq overlay
                            (make-overlay (car match-span) (cdr match-span)))
                      'face (aref face-names-array face-name-index))
                     (overlay-put overlay 'evaporate t)
                     ;; Reference each overlay in order to be able to delete
                     ;; it in case of a new search leading to matches in the
                     ;; same file. It is better than using `overlays-in'
                     ;; because if overlays are used for another goal in the
                     ;; same buffer, we'll be able to delete selectively the
                     ;; ones used to fontify matches.
                     ;; This assignment is local to `buf'.
                     (setq tdbsf-matches-overlays
                           (cons overlay tdbsf-matches-overlays))
                     )
                 (cdr group))
                )
            (cdr record-info))
           )
       matching-records-info-array))))


;;;; Functions to determine the start of the previous or next matching record
;;;; with respect to a buffer position; functions to bring that record under
;;;; the eyes of the user (the reference buffer position defaulting to the
;;;; position of the point).

(defun tdbsf-previous-matching-record-index (pos)
  "Return the global record index of the last matching record preceding POS.
Return the global record index of the last (i.e., highest numbered)
matching record that starts before the buffer position POS.

POS refers to the current buffer but the global record index that is
returned can be that of a record that is not displayed in the current
buffer. In other words, this function can \"traverse files\".

If there is no preceding matching record in the current buffer and
the file it is visiting is the first file that matched, return nil.

This function must be called in a buffer where the local variable
`tdbsf-this-file-s-matching-records-info' has been initialized."
  (let ((i (length tdbsf-this-file-s-matching-records-info))
        (found nil)
        tmp)
    (while (and (not found) (>= (setq i (1- i)) 0))
      (if (<
           (marker-position
            (aref (aref tdbsf-this-file-s-matching-records-info i) 0))
           pos)
          (setq found t)))
    (if found
        (aref (aref tdbsf-this-file-s-matching-records-info i) 1)
      ;; If we didn't find a previous record in this buffer and if the global
      ;; index of the first record of this buffer is 0 (i.e., it is the first
      ;; record of the first file that matched), then return nil.
      (if (zerop (setq tmp (aref (aref
                                  tdbsf-this-file-s-matching-records-info
                                  0) 1)))
          nil
        (1- tmp)))))

(defun tdbsf-next-matching-record-index (pos)
  "Return the global record index of the first matching record after POS.
Return the global record index of the first (i.e., lowest numbered)
matching record that starts after the buffer position POS.

POS refers to the current buffer but the global record index that is
returned can be that of a record that is not displayed in the current
buffer. In other words, this function can \"traverse files\".

If there is no following matching record in the current buffer and
the file it is visiting is the last file that matched, return nil.

This function must be called in a buffer where the local variable
`tdbsf-this-file-s-matching-records-info' has been initialized."
  (let ((i -1)
        (max-loop-count (length tdbsf-this-file-s-matching-records-info))
        (found nil)
        (total-number-of-matching-records
         (car (with-current-buffer tdbsf-summary-buffer-name tdbsf-results)))
        tmp)
    (while (and (not found) (< (setq i (1+ i)) max-loop-count))
      (if (>
           (marker-position
            (aref (aref tdbsf-this-file-s-matching-records-info i) 0))
           pos)
          (setq found t)))
    (if found
        (aref (aref tdbsf-this-file-s-matching-records-info i) 1)
      ;; If we didn't find a next record in this buffer and if the global
      ;; index of the last record of this buffer plus one equals the total
      ;; number of records (i.e., it is the last record of the last file that
      ;; matched), then return nil.
      (if (< (setq tmp (1+ (aref (aref tdbsf-this-file-s-matching-records-info
                                       (1- max-loop-count)) 1)))
             total-number-of-matching-records)
          tmp
        nil))))

(defun tdbsf-bring-record-under-the-eyes (global-record-index buf same-window)
  (let* ((global-records-index-entry
          (aref (with-current-buffer tdbsf-summary-buffer-name
                  tdbsf-matching-records) global-record-index))
         (window-start
         (with-current-buffer buf
           (goto-char (marker-position
                       (aref (aref tdbsf-this-file-s-matching-records-info
                                   (aref global-records-index-entry 2)) 0)))
           (beginning-of-line)
;;; That was an idea, it is not implemented for now.
;;;           (if tdbsf-narrow-to-displayed-record
;;;               (narrow-to-region
;;;                (point) (aref (car (aref global-records-index-entry 1)) 4)))
           (point))))
    (set-window-start
     (if same-window (progn (switch-to-buffer buf) (selected-window))
       (display-buffer buf))
     window-start)))

(defun tdbsf-display-previous-matching-record (position)
  "Display the previous matching record.
Display the previous record that matched the search expression. This is
the last (i.e., highest numbered) matching record that starts before
POSITION in the current buffer.

That record may be in another buffer than the current buffer, or even
in a file that no buffer is visiting yet. In this case, the file will
be visited if necessary and its buffer will be displayed in a window.

If POSITION is in or before the first matching record, print a
message and display the first matching record.

When called interactively, POSITION corresponds to the location of
point."
  (interactive "d")
  (let ((record-index (tdbsf-previous-matching-record-index position)))
    (if record-index
        (tdbsf-display-record record-index t)
      (tdbsf-display-record 0 t)
      (message "Already on the first matching record"))))

(defun tdbsf-display-next-matching-record (position)
  "Display the next matching record.
Display the next record that matched the search expression. This is
the first (i.e., lowest numbered) matching record that starts after
POSITION in the current buffer.

That record may be in another buffer than the current buffer, or even
in a file that no buffer is visiting yet. In this case, the file will
be visited if necessary and its buffer will be displayed in a window.

If POSITION is in or after the last matching record, print a message
and display the last matching record.

When called interactively, POSITION corresponds to the location of
point."
  (interactive "d")
  (let ((record-index (tdbsf-next-matching-record-index position))
        (total-number-of-matching-records
         (car (with-current-buffer tdbsf-summary-buffer-name tdbsf-results))))
    (if record-index
        (tdbsf-display-record record-index t)
      (tdbsf-display-record (1- total-number-of-matching-records) t)
      (message "Already on the last matching record"))))

(define-key tdbsf-db-file-mode-map "\C-c\C-p"
  'tdbsf-display-previous-matching-record)
(define-key tdbsf-db-file-mode-map "\C-c\C-n"
  'tdbsf-display-next-matching-record)

(provide 'tdbsf-db-file-mode)
