;;; tdbsf-summary-mode.el --- TDBSF Summary mode

;; Copyright (c) 2001, 2004, 2009 Florent Rougon

;; Maintainer: Florent Rougon <flo@via.ecp.fr>

;; This file is part of TDBSF.

;; TDBSF is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; TDBSF is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with TDBSF; see the file COPYING. If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(defvar tdbsf-summary-buffer-name "*TDBSF Results Summary*")

(defvar tdbsf-summary-mode-map (make-sparse-keymap)
  "Keymap for TDBSF Summary mode.")

(define-key tdbsf-summary-mode-map "\r" 'tdbsf-display-record-at)

(defvar tdbsf-summary-record-button-keymap
  (let ((map (make-sparse-keymap)))
    (define-key map [mouse-2] 'tdbsf-display-record-at-mouse)
    map)
  "Keymap for record buttons in tdbsf-summary-mode.")

(defvar tdbsf-summary-mode-hook nil
  "*Hook called by `tdbsf-summary-mode'.")

(defface tdbsf-summary-file-name-face
  '((
     ((class color))
     (:foreground "blue"))
    (t (:inverse-video t))              ; not tested
    )
  "Used to display a file name in tdbsf-summary-mode.")

(defun tdbsf-summary-mode ()
  "Major mode used to display and examine the results of a TDBSF search.
\\{tdbsf-summary-mode-map}"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tdbsf-summary-mode-map)
  (mapcar 'make-local-variable
          '(tdbsf-results
            tdbsf-matching-records
            tdbsf-session-id))
  (setq major-mode              'tdbsf-summary-mode
	mode-name               "TDBSF Summary"
	)
  (put 'tdbsf-summary-mode 'mode-class 'special)
  (set (make-local-variable 'truncate-lines) t)
  (run-mode-hooks 'tdbsf-summary-mode-hook)
  )

(defun tdbsf-prepare-db-file-buffer-and-display-it
  (buf record-index matching-records-info-array global-record-info
       same-window)
  ;; Create a marker in `buf' for each record and build an array storing the
  ;; marker and the _global_ index for each record; set
  ;; `tdbsf-matches-overlays' to nil so that references to overlays previously
  ;; used for matches in `buf' are definitely gone; tag as well the buffer
  ;; with a session id so that later tdbsf queries giving different results
  ;; than the last one don't try to reuse these informations.
  (tdbsf-create-this-file-s-records-info-and-markers
   matching-records-info-array buf (- record-index (aref global-record-info 2)))
  ;; Fontify all matches in `buf'
  (tdbsf-fontify-matches matching-records-info-array buf)
  (tdbsf-bring-record-under-the-eyes record-index buf same-window))

(defun tdbsf-display-record (record-index same-window)
  (require 'tdbsf-db-file-mode)
  (let* ((global-record-info (aref
                              (with-current-buffer tdbsf-summary-buffer-name
                                tdbsf-matching-records) record-index))
         (file (car (aref global-record-info 0)))
         (matching-records-info-array (aref (cdr (aref global-record-info 0)) 1))
         ;; Windows semi-case-sensitive filesystems madness
         ;; ***********************************************
         ;;
         ;; The file-truename call is there to avoid the following problem:
         ;;   - in Windows' "filesystems", there does not seem to be a way to
         ;;     get the correct case for a file (as Explorer and an MS-DOS box
         ;;     don't display the same case for a given file)
         ;;   - with glob.glob("d:/foo/*.bar"), the TDBSF engine gets
         ;;     file names with that case preserved (i.e., d:/, foo and bar in
         ;;     lowercase here)
         ;;   - when you open in Emacs a d:/foo/quux.bar file that matched,
         ;;     there is no problem... but `buffer-file-name' in that buffer
         ;;     gets a "canonical" case, which may be different from that of
         ;;     d:/foo/quux.bar
         ;;   - finally, (get-buffer-file "d:/foo/quux.bar") evaluates to nil
         ;;     even if this file is visited in a buffer (on Emacs 21; I am
         ;;     not sure about this on Emacs 20) if the case of
         ;;     "d:/foo/quux.bar" is not the same as in the `buffer-file-name'
         ;;     variable local to that buffer.
          ;;
         ;; With (file-truename "d:/foo/quux.bar"), we get the same case as in
         ;; `buffer-file-name' and there is no problem.
         (buf (get-file-buffer (file-truename file)))
         )
    (unless buf (setq buf (find-file-noselect file)))

    (if (buffer-modified-p buf)
        (message "TDBSF: refusing to use buffer %s, as it is marked as \
modified" (buffer-name buf))

      ;; Put `buf' into tdbsf-db-file-mode if it is not already the case
      (with-current-buffer buf
        (unless (eq major-mode 'tdbsf-db-file-mode)
          (tdbsf-db-file-mode)))

      ;; Is the buffer-local information in `buf' up-to-date?
      (if (and (with-current-buffer buf (boundp 'tdbsf-session-id))
               (with-current-buffer tdbsf-summary-buffer-name
                 (boundp 'tdbsf-session-id))
               (string-equal (buffer-local-value 'tdbsf-session-id buf)
                             (buffer-local-value 'tdbsf-session-id
                                                 (get-buffer
                                                  tdbsf-summary-buffer-name))))
          ;; Yes, reuse it.
          (tdbsf-bring-record-under-the-eyes record-index
                                             buf same-window)
        ;; We don't want to see the matches fontified by previous
        ;; sessions!
        (with-current-buffer buf
          (if (boundp 'tdbsf-matches-overlays)
              (mapcar 'delete-overlay tdbsf-matches-overlays)))
        ;; Warning: at this point, the overlays are "deleted" (i.e.
        ;; detached from `buf' but they are still referenced in
        ;; `tdbsf-matches-overlays'. The following function call ends
        ;; this situation by setting `tdbsf-matches-overlays' to nil.
        (tdbsf-prepare-db-file-buffer-and-display-it
         buf record-index matching-records-info-array global-record-info
         same-window)))))

(defun tdbsf-display-record-at (location)
  "Display the TDBSF record corresponding to the line at LOCATION.

LOCATION is interpreted as a character position in the buffer named
according to `tdbsf-summary-buffer-name'. When called interactively,
LOCATION is set to the value of point."
  (interactive "d")
  (let* ((line                           ; counted starting at 0
         (save-excursion (beginning-of-line) (count-lines 1 (point))))
         (match-count (car
                       (with-current-buffer tdbsf-summary-buffer-name
                         tdbsf-results))))
    (unless (>= line match-count)
      (tdbsf-display-record line nil))))

(defun tdbsf-display-record-at-mouse (event)
  "Display a TDBSF record specified by a mouse click."
  (interactive "e")
  (save-excursion
    (mouse-set-point event)
    (tdbsf-display-record-at (point))))

(provide 'tdbsf-summary-mode)
