;;; tdbsf.el --- calling the TDBSF from Emacs, internal routines

;; Copyright (c) 2001, 2004, 2009, 2013 Florent Rougon

;; Maintainer: Florent Rougon <flo@via.ecp.fr>

;; This file is part of TDBSF.

;; TDBSF is free software; you can redistribute it and/or modify it
;; under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; TDBSF is distributed in the hope that it will be useful, but
;; WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
;; General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with TDBSF; see the file COPYING. If not, write to the Free
;; Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Code:

(require 'tdbsf-init)

(defvar tdbsf-invocation-program-name
  "tdbsf-front-end"
  "*Name of the executable used to run TDBSF.
Typical values are 'tdbsf-front-end' or 'python3'. In the latter case,
the path to 'tdbsf-front-end' must be specified as the first element
of `tdbsf-invocation-args' (if one wants to pass options to the Python
interpreter, it won't be the first element, of course).")

(defvar tdbsf-invocation-args
  '()                          ; [or '("/home/flo/src/tdbsf/tdbsf-front-end")]
  "*List of the first arguments used when invoking the TDBSF engine.
This list contains the first arguments that are passed to the
Python interpreter or TDBSF engine (depending on whether
`tdbsf-invocation-program-name' refers to the Python interpreter
or to the 'tdbsf-front-end' Python program).

The remaining arguments, such as --boh-flags and --search-string
are either hardcoded for this Emacs interface or derived from the
user preferences and the arguments supplied to the
front-end (e.g., \\[tdbsf]).")

(defvar tdbsf-coding-system-for-read 'utf-8
  "Coding system used by Emacs when reading input from tdbsf-front-end.
The value will be bound to `coding-system-for-read' before
calling tdbsf-front-end and should be set in accordance with
`tdbsf-pythonioencoding'.")

(defvar tdbsf-coding-system-for-write 'utf-8
  "Coding system used by Emacs when writing output to tdbsf-front-end.
The value will be bound to `coding-system-for-write' before
calling tdbsf-front-end and should be set in accordance with
`tdbsf-pythonioencoding'.")

(defvar tdbsf-pythonioencoding "utf-8:strict"
  "Coding system used by tdbsf-front-end for I/O operations.
This coding system, and optional error handler, is passed as the
PYTHONIOENCODING environment variable by `call-tdbsf' to
tdbsf-front-end. It affects reading from standard input and
writing to standard output by tdbsf-front-end and should be set
in accordance with `tdbsf-coding-system-for-read' and
`tdbsf-coding-system-for-write'.")

(defvar tdbsf-raw-output-buffer-name "*TDBSF Raw Output*")
(defvar tdbsf-error-buffer-name "*TDBSF Error*")
(defvar tdbsf-temporary-buffer-name " *TDBSF Temporary Data*")

(defvar tdbsf-end-of-emacs-oriented-output-identifier
  "\n\n*** End of TDBSF Emacs-oriented output ***\n"
  "String output by the Emacs interface for the TDBSF after all the useful output.

This string is to ensure we have all the output in the buffer indicated by
`tdbsf-raw-output-buffer-name', because I am not sure that it is always the
case when a process sentinel gets a \"finished\" event.")

(defvar tdbsf-error-identifier
  "TDBSF error:\n\n"
  "String output by the Emacs interface for the TDBSF when it catches an exception.

This string is the first thing that is output. It is then
followed by an explanation of the error.")

(defun tdbsf-insert-record-info (record-elt source-buffer-or-name
                                            &optional extra-info-max-width)
  ;; Note: `extra-info-max-width' does *not* take the separator string
  ;;       inserted before the extra info into account (i.e. here, " ").
  ;;       It is the maximum width *of the extra info*.
  (insert " "
          (with-current-buffer source-buffer-or-name
            (save-restriction
              ;; The record's header
              (narrow-to-region (aref (car record-elt) 1)
                                (aref (car record-elt) 2))
              ;; We do not want the extra info to go beyond `fill-column'
              (if (and extra-info-max-width
                       (< (+ (point-min) extra-info-max-width) (point-max)))
                  (narrow-to-region (point-min)
                                    (+ (point-min) extra-info-max-width)))
              (goto-char (point-min))
              (end-of-line)
              (buffer-substring-no-properties (point-min) (point))))
          "\n"))

(defun tdbsf-process-output ()
  "Process output from the TDBSF engine and display the results in a buffer.
Decode the output supplied by the TDBSF engine, use the resulting
information to construct the buffer named
`tdbsf-summary-buffer-name'. and display this buffer for the user.
Several buffer-local variables are created in that buffer:
  - tdbsf-session-id to avoid problems when several successive
    invocations of TDBSF find matches in the same database files;
  - tdbsf-results to store the full data structure supplied by the
    TDBSF engine;
  - tdbsf-matching-records to store information about all records
    that matched the search expression and, indirectly, about all
    records contained in a file where there was a matching record
    (the latter information being the locations of the start end end
    of the \"boh\" and \"eoh\" parts a record header)."
  (with-current-buffer tdbsf-raw-output-buffer-name
    ;; Check if the tdbsf call was ended by an exception
    (if (string-equal (buffer-substring-no-properties
                         (point-min) (+ (point-min)
                                        (length tdbsf-error-identifier)))
                        tdbsf-error-identifier)
        (progn
          ;; Put the error message in an appropriate buffer
          (let ((start (point-min))
                (end (- (point-max)
                        (length
                         tdbsf-end-of-emacs-oriented-output-identifier))))
            (with-current-buffer
                (get-buffer-create tdbsf-error-buffer-name)
              (erase-buffer)
              (insert-buffer-substring tdbsf-raw-output-buffer-name start end)))
          (display-buffer tdbsf-error-buffer-name)
          ;; Using `error' here causes trouble because the function is called
          ;; by a process sentinel...
          (if (>= emacs-major-version 23)
              (lwarn '(tdbsf) :error "see details in buffer %s"
                     tdbsf-error-buffer-name)
            (message "TDBSF error -- see details in buffer %s"
                     tdbsf-error-buffer-name)))
      (let ((tdbsf-res (read (buffer-string)))
            (global-record-index -1))
        (if (zerop (car tdbsf-res))
            (message "TDBSF: no match found")
          (require 'tdbsf-summary-mode)
          (with-current-buffer
              (get-buffer-create tdbsf-summary-buffer-name)
            (tdbsf-summary-mode)
            ;; Buffer-local bindings
            (setq buffer-read-only nil
                  tdbsf-results tdbsf-res
                  tdbsf-matching-records (make-vector
                                       ;; Total number of records that matched
                                       (car tdbsf-res)
                                       nil)
                  tdbsf-session-id (tdbsf-make-unique-session-id))
            (erase-buffer)
            ;; Loop over the files that matched
            (mapcar
             #'(lambda (file-elt)
                 (let* ((filename (file-name-nondirectory (car file-elt)))
                        overlay start-of-overlay
                        (record-index-in-this-file -1))
                   (with-current-buffer
                       (get-buffer-create tdbsf-temporary-buffer-name)
                     ;; Without the insert-file-contents call below, the
                     ;; following scenario could happen:
                     ;;
                     ;; 1) We just created tdbsf-temporary-buffer-name, it is
                     ;;    empty, not modified.
                     ;; 2) We insert-file-contents an Unix-encoded file
                     ;;    contents in tdbsf-temporary-buffer-name. It gets
                     ;;    Unix encoding, but everything is OK.
                     ;; 3) We insert-file-contents a DOS-encoded file
                     ;;    contents in tdbsf-temporary-buffer-name (with the
                     ;;    REPLACE argument being t). The buffer remains in
                     ;;    Unix encoding, the file contents is inserted with
                     ;;    Carriage Return characters at the end of every line
                     ;;    and as a consequence, the offsets obtained from the
                     ;;    TDBSF engine don't correspond to the offsets in the
                     ;;    Emacs buffer tdbsf-temporary-buffer-name.
                     ;; 4) The records headers for this file are not located
                     ;;    correctly and the summary lines in the TDBSF Summary
                     ;;    buffer are wrong for this file.
                     (set-buffer-file-coding-system 'undecided t)
                     ;; The whole file
                     (insert-file-contents (car file-elt) nil
                                           nil nil t))
                   ;; Loop over the records that matched
                   (mapcar
                    #'(lambda (record-elt)
                        ;; Insert the name of the file corresponding to the
                        ;; current line and remember where it begins
                        (setq start-of-overlay (point))
                        (insert filename)
                        ;; Create and customize an overlay to fontify the file
                        ;; name
                        (overlay-put
                         (setq overlay
                               (make-overlay start-of-overlay (point)))
                         'face 'tdbsf-summary-file-name-face)
                        (overlay-put overlay 'evaporate t)
                        ;; Insert some info about the record corresponding to
                        ;; the current line and buttonize it.
                        (setq start-of-overlay (1+ (point)))
                        ;; No need to pass remaining-width, since we now set
                        ;; `truncate-lines' to t in the summary buffer.
                        (tdbsf-insert-record-info record-elt
                                                  tdbsf-temporary-buffer-name)
                        ;; The buttons are implemented directly with overlays.
                        ;; We could use the "button" library instead, which is
                        ;; shipped with Emacs starting from version 22.
                        (setq overlay
                              (make-overlay start-of-overlay (1- (point))))
                        ;; Set the overlay properties
                        (mapcar #'(lambda (c)
                                    (overlay-put overlay (car c) (cdr c)))
                                `((mouse-face . highlight)
                                  (follow-link . t)
                                  (evaporate . t)
                                  (keymap .
                                          ,tdbsf-summary-record-button-keymap)))
                        ;; These are only *references*, not copies, thus I
                        ;; think there is no resource-eating problem
                        (aset tdbsf-matching-records
                              (setq global-record-index
                                    (1+ global-record-index))
                              (vector file-elt record-elt
                                      (setq record-index-in-this-file
                                            (1+ record-index-in-this-file))))
                        )
                    (aref (cdr file-elt) 1))))
             (cdr tdbsf-res))
            ;; This binding is local to the buffer `tdbsf-summary-buffer-name'
            (setq buffer-read-only t)
            (goto-char (point-min))
            )
          (display-buffer tdbsf-summary-buffer-name)
          )))))

(defun tdbsf-process-sentinel (process event)
  (let ((i -1) (max-iterations 60) input-arrived)
    (when (string-equal event "finished\n")
      (with-current-buffer tdbsf-raw-output-buffer-name
        (while (and (< (setq i (1+ i)) max-iterations)
                    (not input-arrived))
          (if (string-equal
               tdbsf-end-of-emacs-oriented-output-identifier
               (buffer-substring
                (- (point-max)
                   (length tdbsf-end-of-emacs-oriented-output-identifier))
                (point-max)))
              (setq input-arrived t)    ; to end the loop ASAP
            (sleep-for 1)               ; wait a little bit and look again...
            )))
      (if input-arrived
          (tdbsf-process-output)
        (error "Could not see end of TDBSF output"))
      )
    ))

(defun call-tdbsf (expr pattern boh boh-flags eoh eoh-flags search-string
                        header-body-separator db-files-mode
                        &optional crlf-to-lf-in-args)

  ;; Create if necessary and clear the " *TDBSF raw output*" buffer
  (with-current-buffer
      (get-buffer-create tdbsf-raw-output-buffer-name)
    (erase-buffer))

  (let* ((coding-system-for-read tdbsf-coding-system-for-read)
         (coding-system-for-write tdbsf-coding-system-for-write)
         (process-environment
          (cons (format "%s=%s" "PYTHONIOENCODING" tdbsf-pythonioencoding)
                process-environment))
         (arguments
          (append tdbsf-invocation-args
                  (if crlf-to-lf-in-args '("--crlf-to-lf-in-args"))
                  (list
                   "--emacs-friendly-output"
                   "--boh-flags" boh-flags
                   "--eoh-flags" eoh-flags
                   "--db-files-mode" db-files-mode
                   "--search-string" search-string
                   "--header-body-separator" header-body-separator
                   expr pattern boh eoh)))
         (tdbsf-process
          (apply 'start-process "tdbsf"
                 tdbsf-raw-output-buffer-name
                 tdbsf-invocation-program-name
                 arguments)))
    ;; Print the command line for debugging
;     (print (cons tdbsf-invocation-program-name arguments))
    (set-process-sentinel tdbsf-process 'tdbsf-process-sentinel)))

;; The two following functions are borrowed (and very slightly modified) from
;; Gnus v5.8.8, an excellent mail-and-news-reader for Emacs released under the
;; GNU GPL. They are defined in the file message.el of Gnus sources.
;; See http://www.gnus.org/ for more information.

(defun tdbsf-number-base36 (num len)
  (if (if (< len 0)
	  (<= num 0)
	(= len 0))
      ""
    (concat (tdbsf-number-base36 (/ num 36) (1- len))
	    (char-to-string (aref "zyxwvutsrqponmlkjihgfedcba9876543210"
				  (% num 36))))))

(defvar tdbsf-unique-id-char nil)

(defun tdbsf-make-unique-session-id ()
  ;; Don't use microseconds from (current-time), they may be unsupported.
  ;; Instead we use this randomly initialized counter.
  (setq tdbsf-unique-id-char
	(% (1+ (or tdbsf-unique-id-char (logand (random t) (1- (lsh 1 20)))))
	   ;; (current-time) returns 16-bit ints,
	   ;; and 2^16*25 just fits into 4 digits in base 36.
	   (* 25 25)))
  (let ((tm (current-time)))
    (concat
     (if (memq system-type '(ms-dos emx vax-vms))
	 (let ((user (downcase (user-login-name))))
	   (while (string-match "[^a-z0-9_]" user)
	     (aset user (match-beginning 0) ?_))
	   user)
       (tdbsf-number-base36 (user-uid) -1))
     (tdbsf-number-base36 (+ (car   tm)
                             (lsh (% tdbsf-unique-id-char 25) 16)) 4)
     (tdbsf-number-base36 (+ (nth 1 tm)
                             (lsh (/ tdbsf-unique-id-char 25) 16)) 4)
     (number-to-string (random t)))))

;; Based on `emacs-version'
(defun tdbsf-version (&optional here) "\
Return string describing the version of TDBSF that is installed.
If optional argument HERE is non-nil, insert string at point."
  (interactive "P")
  (let ((version-string
         (format "%s version %s" "TDBSF" tdbsf-version)))
    (if here
        (insert version-string)
      (if (called-interactively-p 'interactive)
          (message "%s" version-string)
        version-string))))

(provide 'tdbsf)

;; Local Variables:
;; coding: utf-8
;; End:
