# expr2pfe.py --- Parser module for a TDBSF search expression
# -*- coding: utf-8 -*-

# Copyright (c) 2001, 2004, 2009, 2013 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.


import sys, string, io
from .stack_impl import Stack
from . import error as TDBSFGenericException


# The only exception defined by this module
class TDBSFSyntaxError(TDBSFGenericException):
    ExceptionShortDescription = "Syntax error in match expression"

global elts


class Output:
    def write(self, s):
        sys.stdout.write(s)


class Parser:
    """Parser for TDBSF search expressions.

    The expression is read from a file-like object passed as the
    first argument to the constructor.

    The parsing is started by calling the method gen_pfe(), which
    reads until EOF and returns a list called PFE (PostFix
    Expression) that describes the search expression in Reverse
    Polish Notation (RPN). In previous versions, the PFE was called
    CPF, standing for "Chaîne PostFixée" (French).

    The parser does not close the input stream, it is the
    responsibility of the caller to do so.

    """
    def __init__(self, input_stream):
        self.input = input_stream
        self.line = 1

    def err(self, s):
        raise TDBSFSyntaxError("%s (line %d)" % (s, self.line))

    def expr(self):
        self.term()
        while self.reco("|"): self.term()

    def term(self):
        self.factor()
        while self.reco("&"): self.factor()

    def factor(self):
        if self.reco("("):
            self.expr()
            if not self.reco(")"):
                self.err("missing close-parenthesis")
        elif self.reco("!"):
            self.factor()
        else:
            self.atom()

    def atom(self):
        self.regexp()
        while self.reco("~"):
            self.integer()
            self.atom()

    # Does NOT tolerate whitespace at the beginning
    def integer(self):
        l = []

        while True:
            pos, c = self.input.tell(), self.input.read(1)
            if not c: break
            elif c in string.digits:
                l.append(c)
            else:
                # This is illegal in Python 3 with text streams, although it
                # happened to work with StringIO in Python 2 (because relative
                # seeking into a string, Unicode or not, is easy).
                #   self.input.seek(-1,1)
                self.input.seek(pos, io.SEEK_SET)
                break
        if l:
            self.pfe.append("".join(l))
        else:
            self.err("'~' must be immediately followed by an integer")

    def regexp(self):
        pattern_str = []
        flags_str = []

        # The beginning of the read string defines a pattern. Then,
        # if this string is a quoted one and contains a non-escaped
        # '%', we'll get the flags the user wants to apply with the
        # pattern.
        l = pattern_str

        while True:
            c = self.input.read(1)
            if not c: self.err("regular expression expected at end of file")
            elif c in ' \t': pass     # ignore spaces and tabs before regexps
            elif c == '\n': self.line += 1       # as well as newlines
            else:
                break

        if c == '"': # Quoted regexp
            while True:
                c = self.input.read(1)
                if not c: self.err("quoted regexp not terminated at end of file")
                elif c == '\\':             # escape character
                    c = self.input.read(1)
                    if not c:
                        self.err("lonely '\\' in quoted regexp at end of file")
                    else:
                        l.append(c)
                elif c == '"':
                    break
                elif c == '%': # Next characters belong to the "flags" part
                    l = flags_str
                elif c == '\n':
                    self.line += 1
                    l.append(c)
                else: l.append(c)
        elif c in "()|&!~": # No regexp begins with such chars
            self.err("expected a regexp, but found '%s'" % c)
        else: # Unquoted regexps may contain raw '\' and '"' characters
            l.append(c) # We have already read the regexp's 1st character
            while True:
                pos, c = self.input.tell(), self.input.read(1)
                if not c: break
                elif c in " \t\n)|&~": # These chars mark the end of an unquoted
                                       # regexp
                    self.input.seek(pos, io.SEEK_SET)
                    break
                elif c in "(!": self.err("regexp terminated by '%s'" % c)
                else: l.append(c)

        # 'r' indicates a regexp
        self.pfe.append(('r', "".join(pattern_str), "".join(flags_str)))

    def handle(self, elt_name, attrib=None):
        # Opening elements are put on the top of the stack
        if elts[elt_name].opening:
            self.stack.push(Elt(elt_name, attrib))
        else:
            # Conditional pop
            while elts[elt_name].prio <= elts[self.stack[-1].name].prio:
                self.pop_elt()
            # Closing element? (e.g., closing parenthesis)
            if elts[elt_name].closing:
                stack_top = elts[self.stack[-1].name]
                # Yes. Is the matching opening element present?
                if stack_top.opening and (stack_top.opening == elt_name):
                    self.pop_elt()
                else:
                    self.err("got '%s' without matching '%s'"
                        %(elt_name, elts[elt_name].closing))
            else:
                self.stack.push(Elt(elt_name, attrib))

    # Pop an element from the stack and update the PFE
    def pop_elt(self):
        pfe_code = elts[self.stack.pop().name].pfe
        if pfe_code:
            self.pfe.append(pfe_code)

    def reco(self, s):
        while True:
            pos, c = self.input.tell(), self.input.read(1)
            if not c: return 0
            elif c in " \t": pass           # ignore spaces and tabs before tokens
            elif c == '\n': self.line += 1       # as well as newlines
            else:
                self.input.seek(pos, io.SEEK_SET)
                break

        n = len(s)
        # WARNING: if 's' contains newlines, the caller must increase
        # self.line if 'self.reco' returns 1.
        pos, filestr = self.input.tell(), self.input.read(n)
        if s == filestr:
            self.handle(s)
            return 1
        else:
            self.input.seek(pos, io.SEEK_SET)
            return 0

    def gen_pfe(self):
        self.stack = Stack()
        self.stack.push(Elt("BOF")) # Beginning Of File
        self.pfe = []

        self.expr()

        while True:
            c = self.input.read(1)
            if not c:
                self.handle("EOF")
                break
            elif c in " \t": pass
            elif c == '\n': self.line += 1
            else: self.err("data found after a valid expression")

        return self.pfe


class EltProps:
    def __init__(self, name, pfe, prio, opening, closing):
        self.name = name
        self.pfe = pfe
        self.prio = prio
        self.opening = opening
        self.closing = closing

    def __str__(self):
        return "<EltProps: name : '" + self.name \
              + "', pfe : '" + self.pfe \
              + "', prio : '" + str(self.prio) \
              + "', opening : '" + str(self.opening) \
              + "', closing : '" + str(self.closing) + "'>"


class Elt:
    def __init__(self, name, attrib=None):
        self.name = name
        self.attrib = attrib or {}

    def __str__(self):
        return "<Elt: '" + self.name + "', " + str(self.attrib) + ">"


def init_elts():
    global elts
    elts = { 'BOF': EltProps("BOF", "", 0, "EOF", 0),
             '(': EltProps("(", "", 0, ")", 0),
             'EOF': EltProps("EOF", "", 1, 0, "BOF"),
             ')': EltProps(")", "", 1, 0, "("),
             '|': EltProps("|", "|", 2, 0, 0), # OR
             '&': EltProps("&", "&", 3, 0, 0), # AND
             '!': EltProps("!", "!", 4, 0, 0), # NOT
             '~': EltProps("~", "~", 5, 0, 0)  # NEAR
             }


def gen_pfe_from_file(filename, *args, **kwargs):
    return Parser(open(filename, 'r', *args, **kwargs)).gen_pfe()

def gen_pfe_from_string(s):
    return Parser(io.StringIO(s)).gen_pfe()


def test(root_dir, quiet=False):
    """Quick regression test of this module.

    This test does not require any external file. It can be run from the root
    directory of the TDBSF distribution with the command:

      python3 -m tdbsf.expr2pfe

    """
    try:
        pfe = gen_pfe_from_string(r'''élan & (bsdqqsd |!(c ~32 h))    & defg\y |
"Ô cba\"g\\\%e()~%I|L|X" &
  (ghi ~3 abc ~4 def ~5 "kjh()!") | "d\e\" maraîchère"''')
    except TDBSFGenericException as exc_instance:
        print(exc_instance.complete_message(), file=sys.stderr)
        return 1

    assert pfe == [('r', 'élan', ''), ('r', 'bsdqqsd', ''), ('r', 'c', ''), '32',
                   ('r', 'h', ''), '~', '!', '|', '&', ('r', 'defg\\y', ''),
                   '&', ('r', 'Ô cba"g\\%e()~', 'I|L|X'), ('r', 'ghi', ''), '3',
                   ('r', 'abc', ''), '~', '4', ('r', 'def', ''), '~', '5',
                   ('r', 'kjh()!', ''), '~', '&', '|', ('r', 'de" maraîchère',
                                                        ''), '|']

    if not quiet:
        output = Output()
        output.write(str(pfe) + '\n')

    return 0


init_elts()

if __name__ == "__main__":
    try:
        root_dir = sys.argv[1]
    except IndexError:
        root_dir = ""

    retcode = test(root_dir)
    assert isinstance(retcode, int), retcode
    sys.exit(retcode)
