# pfe2tree.py --- Tree building and optimization for already parsed TDBSF
#                 search expressions
# Copyright (c) 2001, 2004, 2009, 2013 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.


import sys, re, string
from .stack_impl import Stack
from . import expr2pfe
from . import error as TDBSFGenericException


# The only exception defined by this module
class TDBSFRegexpCompilationError(TDBSFGenericException):
    ExceptionShortDescription = "Regular expression compilation error"


def _elt1(s):
    return s[1]

def one_op_and_one_regexp(a, b, stack):
    a[2].append(b)                      # a[2] is modified "in place"
    a[2].sort(key=_elt1)                # a[2] is modified "in place"
    stack.push([a[0], a[1] + b[1] + 2, a[2]]) # + 2 is somewhat arbitrary!


def build_node_sort(a, b, stack, op):
    heaviness = a[1] + b[1] + 2 # + 2 is somewhat arbitrary
    if a[1] <= b[1]:
        stack.push([op, heaviness, [a, b]])
    else:
        stack.push([op, heaviness, [b, a]])


def compile_regexp(pattern, flags_string):
    flags = 0
    for flag in flags_string.split('|'):
        if not flag:
            continue
        try:
            flags |= getattr(re, flag)
        except AttributeError:
            raise TDBSFRegexpCompilationError(
                "no such attribute in the 're' module: '{0}'".format(flag))

    try:
        regexp_object = re.compile(pattern, flags)
    except re.error as e:
        raise TDBSFRegexpCompilationError(str(e))

    return regexp_object


def build_regexp_leaf(pfe_elt): # "leaf" of the "tree" of operations
    pattern = pfe_elt[1]
    # The "heaviness" of the regexp is set to the length of its pattern.
    return ['r', len(pattern), compile_regexp(pattern, pfe_elt[2])]


def pfe2tree(pfe):
    stack = Stack()

    i = 0
    while i < len(pfe):
        elt = pfe[i]
        op = elt[0] # operator like "&" or "!", or regexp indicated by "r"
        if op == 'r':
            stack.push(build_regexp_leaf(elt))
        elif op in "|&":
            a = stack.pop()
            b = stack.pop()
            if (a[0] == op) and (b[0] == 'r'):
                one_op_and_one_regexp(a, b, stack)
            elif (b[0] == op) and (a[0] == 'r'):
                one_op_and_one_regexp(b, a, stack)
            elif a[0] == b[0] == op:
                a[2].extend(b[2])       # a[2] is modified "in place"
                a[2].sort(key=_elt1)    # a[2] is modified "in place"
                                        # + 2 is somewhat arbitrary here:
                stack.push([a[0], a[1] + b[1] + 2, a[2]])
            else: # We cannot group the operations and their arguments
                build_node_sort(a, b, stack, op)
        elif op == '~':
            b = stack.pop()             # is a "regexp leaf"
            n = int(stack.pop())        # is an integer
            a = stack.pop()             # is a "regexp leaf"

            # Each regexp in a "near expression" increases a lot the
            # "heaviness" of the expression; +1 are here so that regexps
            # with "heaviness" 1 don't let the "heaviness" of the "near
            # expression" unchanged.
            c = ['~', (a[1]+1)*(b[1]+1), [a, b], [n]]

            # Get all the "near" expression at once. This enables to optimize
            # it afterwards by an optional inversion.
            while (i + 4 <= len(pfe)) \
                  and (pfe[i+1][0] in string.digits) \
                  and (pfe[i+2][0] == 'r') \
                  and (pfe[i+3][0] == '~'):
                new_regexp = build_regexp_leaf(pfe[i+2])
                # Each regexp in a "near expression" increases a lot the
                # "heaviness" of the expression; +1 are here so that regexps
                # with "heaviness" 1 don't let the "heaviness" of the "near
                # expression" unchanged.
                c[1] = c[1]*(new_regexp[1]+1)
                c[2].append(new_regexp)
                c[3].append(int(pfe[i+1]))
                i += 3

            # The resulting c "leaf" of the tree has the form:
            # ["~", xyz, [ [], [],..., [], []],
            #            [   n1, n2,..., np-1]]
            # where the p "[]" on the first line are "regexp leafs" (see
            # 'build_regexp_leaf' if you want to known what they look like).

            # Optimization: the last regexp should be *less* heavy than the
            # first one in the case of a "r1 ~n1 r2 ~n2 ... ~np-1 rp"
            # expression.
            if c[2][0][1] < c[2][-1][1]:
                c[2].reverse() # invert the order of the p regexps
                c[3].reverse() # invert the order of n1, n2, ..., np-1

            stack.push(c)               # Put our final "near leaf" on the
                                        # stack.
        elif op == '!':
            a = stack.pop()
            stack.push(['!', a[1] + 2, a]) # The + 2 is somewhat arbitrary
        else: # elt should be a string of digits (the n or p in 'a ~n b ~p c')
            stack.push(elt)

        i += 1

    return stack.pop()


def test(root_dir, quiet=False):
    """Quick regression test of this module.

    The test requires a specific search expression located at
    <root_dir>/tests/search_expr. This can be satisfied if 'root_dir'
    is a possibly empty, absolute or relative path to the root
    directory of the TDBSF distribution.

    From the shell command line, assuming the current directory is
    the root directory of the TDBSF distribution, the test can be run
    with:

      python3 -m tdbsf.pfe2tree

    """
    import os
    retcode = 1
    try:
        tree = pfe2tree(
            expr2pfe.gen_pfe_from_file(
                os.path.join(root_dir, "tests", "search_expr"),
                encoding="utf-8"))
    except TDBSFGenericException as exc_instance:
        print(exc_instance.complete_message(), file=sys.stderr)
    else:
        if not quiet:
            print(tree)
        retcode = 0

    return retcode


if __name__ == "__main__":
    try:
        root_dir = sys.argv[1]
    except IndexError:
        root_dir = ""

    retcode = test(root_dir)
    assert isinstance(retcode, int), retcode
    sys.exit(retcode)
