# -*- coding: utf-8 -*-

# tdbsf_api.py --- Module performing the searches through TDBSF database files
# Copyright (c) 2001, 2004, 2009, 2013 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""API for the Trivial DataBase Search Facility.

This module provides a Crawler class that has a 'search' method. This
method can be used to find records that match a given search
expression among a set of files conforming to a TDB format.

"""

import string, re, glob, os, sys, unicodedata, codecs, traceback, textwrap
from textwrap import dedent

from . import expr2pfe
from . import pfe2tree
from . import error as TDBSFGenericException


tw = textwrap.TextWrapper(width=80, break_long_words=False,
                          break_on_hyphens=False)

# Regexps used for parsing the start of db file in binary mode
emacs_style_file_vars_cre = re.compile(
    br"(.* )?-\*- (?P<assignments>.*) -\*-")
emacs_style_file_vars_assignment_cre = re.compile(
    br"^ *(?P<name>[^ \t:]+): (?P<value>[^ \t]+) *$")
binary_newline_cre = re.compile(br"[\r\n]|\r\n")


# Exceptions defined by this module
class TDBSFInvalidDBFileFormatError(TDBSFGenericException):
    ExceptionShortDescription = "Invalid database file format"

class TDBSFInvalidWhereArg(TDBSFGenericException):
    ExceptionShortDescription = "Invalid 'where' argument to Crawler.search()"

class TDBSFInvalidNearArg(TDBSFGenericException):
     ExceptionShortDescription = 'Invalid "near" (\'~\') argument'

class TDBSFInvalidFileHeader(TDBSFGenericException):
     ExceptionShortDescription = 'Invalid file header'

class TDBSFFileDecodingError(TDBSFGenericException):
     ExceptionShortDescription = 'Error decoding file'

# The following characters will be considered word-constituent for near
# expressions (e.g. 'foo ~3 "bars+" ~5 quux').
def is_word_char(c):
    # I don't know much about the Unicode Character Database, but this is
    # probably better than the restrictive western-european centric set
    # commented below.
    # L: letter
    # N: number
    return (c in "-_") or unicodedata.category(c)[0] in 'LN'
# Old method: (note that it didn't contain string.digits)
# word_chars = string.ascii_letters + \
#              "-_ààâäéèêëíìîïóòôöúùûüýÿÀÀÂÄÉÈÊËÍÌÎÏÓÒÔÖÚÙÛÜÝŸ"


def locate_records(filename, filecontents, boh_re_obj, eoh_re_obj):
    """Locate TDB-type records in a string.

    The records are located in string 'filecontents' using compiled regexps
    objects 'boh_re_obj' and 'eoh_re_obj'.

    """

    res = []
    index = 0

    while True:
        boh_match_object = boh_re_obj.search(filecontents, index)
        if not boh_match_object:
            break

        eoh_match_object = eoh_re_obj.search(filecontents,
                                             boh_match_object.end())
        if not eoh_match_object:
            raise TDBSFInvalidDBFileFormatError(
                  "the record beginning at position {0} in '{1}' has "
                  "no 'end of header' delimiter".format(
                    boh_match_object.start(), filename))

        res.append((boh_match_object.span(), eoh_match_object.span()))
        index = eoh_match_object.end()

    return res


def try_regexp_matches(s, lower_limit, upper_limit, regexps,
                       max_offsets):
    while lower_limit < upper_limit:
        match = regexps[0][2].search(s, lower_limit)
        if match:
            # Absolute positions in 's' of the match start and end
            match_span = match.span()
            # Did the match start at a "legal" position?
            if (match_span[0] < upper_limit):
            # If max_offsets is [], 'regexps' should have only one element,
            # the last regexp of the near expression, which has just matched.
               if not max_offsets:
                   return [match_span]
               # Go further trough the "near chain" of regexps
               next_res = near_next_regexp(s, match_span[0],
                                           regexps[1:], max_offsets)
               if next_res:
                   next_res.append(match_span)
                   return next_res
            # Look for another match for this regexp beginning further than
            # this one.
            lower_limit = match_span[0] + 1
        else:
            break

    return 0                            # The near expression didn't match


def near_next_regexp(s, anchor, regexps, max_offsets):
    # If anchor is in a word, put it at the beginning of that word.
    if is_word_char(s[anchor]):
        while (anchor > 0) and is_word_char(s[anchor-1]):
            anchor = anchor - 1

    lower_limit = anchor
    for i in range(max_offsets[0]):
        while (lower_limit > 0) and not is_word_char(s[lower_limit-1]):
            lower_limit = lower_limit - 1
        while (lower_limit > 0) and is_word_char(s[lower_limit-1]):
            lower_limit = lower_limit - 1
        if lower_limit == 0:            # Don't do useless loops
            break

    upper_limit = anchor
    while (upper_limit < len(s)) \
          and is_word_char(s[upper_limit]):
        upper_limit = upper_limit + 1
    for i in range(max_offsets[0]):
        while (upper_limit < len(s)) \
              and not is_word_char(s[upper_limit]):
            upper_limit = upper_limit + 1
        while (upper_limit < len(s)) \
              and is_word_char(s[upper_limit]):
            upper_limit = upper_limit + 1
        if upper_limit == len(s):
            break                       # Don't do useless loops

    return try_regexp_matches(s, lower_limit, upper_limit,
                              regexps, max_offsets[1:])


def test_expr(test_tree, s):
    """Test if there is match of 'test_tree' in 's'."""
    op = test_tree[0] # operator, or 'r' indicating a regexp
    if op == '|':
        for or_arg in test_tree[2]:
            res = test_expr(or_arg, s)
            if res:
                return res
        return 0
    elif op == '&':
        res = []
        for and_arg in test_tree[2]:
            and_arg_res = test_expr(and_arg, s)
            if not and_arg_res:
                return 0
            # ! operations return only 0 or 1; they don't return any match span
            # information.
            elif and_arg_res != 1:
                res.append(and_arg_res)
        return ['&', res]
    elif op == '!':
        return not test_expr(test_tree[2], s)
    elif op == '~':
        regexps = test_tree[2]
        max_offsets = test_tree[3]
        # Verify user input
        for i in max_offsets:
            if i < 1:
                raise TDBSFInvalidNearArg(
                    "each word count 'n_i' in a near expression "
                    "'a_1 ~n_1 a_2 ~n_2 ... ~n_p a_(p+1)' must be strictly "
                    "positive")
        # Recurse to evaluate the near expression
        res = try_regexp_matches(s, 0, len(s),
                                 regexps, max_offsets)
        if res:
            return ['~', res]
        else:
            return 0

    else:
        assert op == 'r', op    # regexp
        # Beware: this will return a MatchObject instance or None, *not* 1 or
        # 0.
        res = test_tree[2].search(s)
        if res:
            return res.span()
        else:
            return 0


# def try_to_guess_eol_type(str):
#     # eol = End Of Line
#     max_bytes_check_count = 4000
#     max_eol_check_count = 3

#     beg = str[:max_bytes_check_count]

#     # Try to find the end-of-line style (CR, LF or CRLF) using the same
#     # algorithm as in GNU Emacs 21.1 (file "coding.c").
#     eol_type = "undecided"
#     eols_found = 0
#     cr_or_lf_re = re.compile("[\r\n]")
#     index = 0

#     while (eols_found < max_eol_check_count):
#         substring = beg[index:]
#         mo = cr_or_lf_re.search(substring)

#         if (mo is None): break          # nothing found until the end of the
#                                         # string
#         ms = mo.start()
#         me = mo.end()

#         index = index + me
#         eols_found = eols_found + 1
#         if substring[ms:me] == '\n':
#             this_eol_type = "LF"
#         elif ((me == len(substring)) or substring[ms+1:ms+2] != '\n'):
#             this_eol_type = "CR"
#         else:
#             this_eol_type = "CRLF"
#             index = index + 1       # Skip the following '\n'

#         if (eol_type == "undecided"):
#             # This is the first end-of-line
#             eol_type = this_eol_type
#         elif (this_eol_type != eol_type):
#             eol_type = "undecided"
#             break
#     return eol_type


# def try_to_convert_eols_to_lfs(s):
#     eol_type = try_to_guess_eol_type(s)

#     if (eol_type == "CR"):
#         return s.replace('\r', '\n')
#     elif (eol_type == "CRLF"):
#         return s.replace('\r\n', '\n')
#     else:
#         # Either LF or inconsistent end-of-lines or no eol found within the
#         # first characters of the file
#         return s

def _resfv_set_fallback_encoding(dict_, encoding):
    if dict_["coding"] is None:
        dict_["coding"] = encoding


def read_emacs_style_file_variables(filename):
    """Read an optional Emacs-style line declaring file variables.

    The line should have the syntax:

      -*- VAR1: VALUE1; VAR2: VALUE2; ... -*-

    (the last semicolon in the declaration is optional)

    For now, only the "coding" pseudo-variable is used by TDBSF. The
    associated value should be a valid encoding name for Python (and
    also for Emacs if you use it to open the files). The encoding
    used must be ASCII-compatible to allow parsing of the
    pseudo-variable line where it is declared (UTF-8 and most
    single-byte encodings have this property). Other variables of the
    pseudo-variable line are ignored as of now.

    Valid encoding names include "utf-8", "iso-8859-1",
    "iso-8859-15". One can check the validity of an encoding name
    with:

      python -c "import codecs; print(codecs.lookup('encoding_name'))"

    If it prints something like:

      <codecs.CodecInfo object for encoding utf-8 at 0x7f68f0c28ae0>

    this means 'encoding_name' is a valid encoding name for Python.

    The declaration line should be immediately followed by a blank
    line, then the first record header, though this is not enforced
    for now.

    FILENAME -- file to parse (an absolute or relative path)

    This function returns a dictionary with a key for every
    Emacs-style variable. The "coding" variable receives special
    treatment and is guaranteed to contain a valid Python encoding
    name upon return.

    """
    res = { "coding": None }
    fallback_encoding = "ascii"

    # First check for a UTF-8 BOM
    with open(filename, mode="rb") as f:
        start = f.read(3)
        if start == b'\xef\xbb\xbf':
            res["coding"] = "utf-8"
        else:
            f.seek(0)

        # Look for an encoding declaration in the first line of the file. Text
        # mode cannot be used because we don't know the encoding yet.
        stat_struct = os.fstat(f.fileno())
        try:
              blksize = stat_struct.st_blksize
        except AttributeError:
              blksize = 4096    # somewhat arbitrary, but large enough

        s = f.read(blksize)

    nlmo = binary_newline_cre.search(s)
    if not nlmo:
        _resfv_set_fallback_encoding(res, fallback_encoding)
        return res

    line = s[0:nlmo.start()]
    mo = emacs_style_file_vars_cre.match(line)

    if not mo:
        _resfv_set_fallback_encoding(res, fallback_encoding)
        return res

    assignments = mo.group("assignments").split(b";")
    if not assignments[-1]:
        assignments = assignments[:-1]

    for a in assignments:
        assignment_mo = emacs_style_file_vars_assignment_cre.match(a)
        if not assignment_mo:
            raise TDBSFInvalidFileHeader(
                "'{0}': invalid assignment in Emacs-style file variables "
                "line: '{1}'".format(
                    filename,
                    str(a, encoding=(res["coding"] or fallback_encoding),
                        errors='replace')))

        name_b = assignment_mo.group("name")
        if name_b == b"coding":
            try:
                encoding = assignment_mo.group("value").decode(
                    res["coding"] or fallback_encoding)
            except UnicodeDecodeError as e:
                raise TDBSFFileDecodingError("'{0}': {1}".format(filename, e))

            try:
                encoding_canon = codecs.lookup(encoding).name
            except LookupError:
                raise TDBSFInvalidFileHeader(
                    "'{0}': encoding is not recognized by Python: {1}".format(
                        filename, encoding))

            if (res["coding"] is not None) and \
                    codecs.lookup(res["coding"]).name != encoding_canon:
                raise TDBSFInvalidFileHeader(
                "'{0}': conflicting declaration between the UTF-8 Byte "
                "Order Mark and the Emacs-style 'coding' file variable "
                "(value: '{1}')".format(filename, encoding))

            # This is a (Unicode) string which is a valid encoding name
            # for the codecs module
            res["coding"] = encoding
        else:
            # res["coding"] should be valid for decoding at this point
            try:
                name, value = (
                    assignment_mo.group(field).decode(res["coding"])
                    for field in (name, value) )
            except UnicodeDecodeError as e:
                raise TDBSFFileDecodingError("'{0}': {1}".format(filename, e))

            res[name] = value

    return res


class Crawler:
    def __init__(self, pattern, boh, boh_flags, eoh, eoh_flags, *,
                 db_files_mode='r', record_positions_offset=0):
        """Initialize a TDBSF Crawler instance.

        A Crawler instance defines a TDBSF database, how to open its
        files and allows searching through these files.

        The files composing the database are identified by 'pattern',
        which is passed to the glob module. The syntax for 'pattern'
        is therefore close to that of Unix shells.

        Each file of the database is interpreted as ASCII, unless it
        starts with an Emacs-style pseudo-comment that declares a
        different encoding (cf. read_emacs_style_file_variables()).
        It must contain records that are located with two regular
        expressions that define a header and a body for each record.

        'boh' is the regular expression used to identify the
        beginning of a header. Its counterpart, 'eoh', marks the end
        of a header and the start of the body of the same record. The
        'boh' and 'eoh' parts of a record are excluded from the
        actual text where the search expression is evaluated when
        using the search() method.

        The 'boh' and 'eoh' regular expressions are interpreted by
        the Python 're' module
        (http://docs.python.org/3/library/re.html) with flags
        respectively defined by 'boh_flags' and 'eoh_flags'. Each of
        these strings should contain the desired flag names separated
        by a '|' character.

        Examples: "IGNORECASE", or equivalently "I"
                  "MULTILINE|VERBOSE|IGNORECASE", or equivalently
                  "M|X|I"

        'db_files_mode' is the mode string (e.g., "r") to use by
        Python's open() built-in function when opening database files
        (after the initial parsing of Emacs-style variables). It must
        be a text mode that allows reading. Don't specify it unless
        you really need to change the default. The database files are
        not opened when creating a Crawler instance; they are opened
        and read whenever the search() method is called.

        The 'record_positions_offset' parameter is used by the
        search() method: it is a global offset added to every file
        position (such as the character offset where a record starts
        or ends in a given file) in the object returned by search().
        It is useful to adapt to 0-based or 1-based character
        indexing.

        """

        l = locals()
        for param in ("pattern", "boh", "boh_flags", "eoh", "eoh_flags",
                      "db_files_mode", "record_positions_offset"):
            setattr(self, param, l[param])

        # Will be initialized by the appropriate function
        self.number_of_matches = None

    def _add_to_result(self, res, records, record_index, match_place, match_pos,
                       offset):
        """Update 'res' after a match of the search expression.

        Update 'res' indicating a match for the record identified by
        'record_index' in 'records'.

        XXX incomplete

        """
        self.number_of_matches += 1

        res.append((
            (
            # Beginning of the header of the matching record (just before
            # the "boh" part)
            records[record_index][0][0] + offset,
            # Beginning of the header of the matching record (just after
            # the "boh" part)
            records[record_index][0][1] + offset,
            # Enf of the header of the matching record (just before the
            # "eoh" part)
            records[record_index][1][0] + offset,
            # Beginning of the record's body (just after the header and
            # its "eoh" part)
            records[record_index][1][1] + offset,
            # End of the matching record (i.e., beginning of the next
            # record or end of file if the matching record is the file's
            # last record)
            records[record_index+1][0][0] + offset
            ),
            (
            # Matched regexps positions (primarily intended for
            # fontification purposes)
            match_place, match_pos
            )
            ))

    def search(self, expr, where="header_plus_body", header_body_separator="¤"):
        """Search for matches of a TDBSF search expression among a set of files.

        Search for matches of 'expr' in the TDBSF database defined
        when the __init__() constructor was called.

        'where' specifies, for each record, the string where 'expr'
        is to be evaluated. It must have one of the following values:
          - "header"           to evaluate 'expr' in the header of
            each record
          - "body"             to evaluate 'expr' in the body of
            each record
          - "header_or_body"   to evaluate 'expr' sequentially in the
            header and in the body of each record (a record is
            retained if 'expr' matched either in its header or in its
            body)
          - "header_plus_body" to evaluate 'expr' in a string which
            is, for each record, the concatenation of its header, the
            value of 'header_body_separator' (which should be a
            string) and its body.

        Return an object describing all matches found. The character
        offsets inside that object can be globally shifted by passing
        a non-zero 'record_positions_offset' parameter to the class
        constructor.

        XXX tell precisely what is returned

        """
        self.number_of_matches = 0

        # Regular expression objects for the beginning of header and
        # end of header identification.
        boh_re_obj = pfe2tree.compile_regexp(self.boh, self.boh_flags)
        eoh_re_obj = pfe2tree.compile_regexp(self.eoh, self.eoh_flags)

        res = []

        for filename in glob.iglob(self.pattern):
            file_vars = read_emacs_style_file_variables(filename)

            with open(filename, self.db_files_mode,
                      encoding=file_vars["coding"]) as f:
                try:
                    filecontents = f.read()
                except UnicodeDecodeError as e:
                    raise TDBSFFileDecodingError("'{0}': {1}".format(filename,
                                                                     e))

            records = locate_records(filename, filecontents,
                                     boh_re_obj, eoh_re_obj)
            # To make the next loop easier
            records.append(((len(filecontents), 0), (0,0)))

            test_tree = pfe2tree.pfe2tree(expr2pfe.gen_pfe_from_string(expr))

            # Results concerning the currently examined database file
            file_res = []

            # The loop below is repeated _inside_ the tests' clauses for
            # efficiency reasons.
            if where == "header":
                # - 1 because we added an artificial element to 'records'
                for i in range(len(records) - 1):
                    match_pos = test_expr(test_tree,
                                          filecontents[records[i][0][1]
                                                       :records[i][1][0]])
                    if match_pos:
                        self._add_to_result(file_res, records, i, "header", match_pos,
                                      self.record_positions_offset)
            elif where == "body":
                # - 1 because we added an artificial element to 'records'
                for i in range(len(records) - 1):
                    match_pos = test_expr(test_tree,
                                          filecontents[records[i][1][1]
                                                       :records[i+1][0][0]])
                    if match_pos:
                        self._add_to_result(file_res, records, i, "body", match_pos,
                                      self.record_positions_offset)
            elif where == "header_or_body":
                # - 1 because we added an artificial element to 'records'
                for i in range(len(records) - 1):
                    match_pos = test_expr(test_tree,
                                          filecontents[records[i][0][1]
                                                       :records[i][1][0]])
                    if match_pos:
                        match_place = "header"
                    else:
                        match_pos = test_expr(test_tree,
                                              filecontents[records[i][1][1]
                                                           :records[i+1][0][0]])
                        match_place = "body"
                    if match_pos:
                        self._add_to_result(file_res, records, i, match_place,
                                      match_pos, self.record_positions_offset)
            elif where == "header_plus_body":
                # - 1 because we added an artificial element to 'records'
                for i in range(len(records) - 1):
                    match_pos = test_expr(test_tree,
                                          filecontents[records[i][0][1]
                                                       :records[i][1][0]]
                                          + header_body_separator +
                                          filecontents[records[i][1][1]
                                                       :records[i+1][0][0]])
                    if match_pos:
                        self._add_to_result(file_res, records, i, "header_plus_body",
                                      match_pos, self.record_positions_offset)
            else:
                raise TDBSFInvalidWhereArg(
                  '\'where\' should be one of "header", "body", '
                  '"header_or_body" and "header_plus_body"')

            if file_res:                    # if a record in this file matched
                res.append([os.path.normpath(filename), records[:-1], file_res])
        return (self.number_of_matches, res)


def lookup_expr(expr, pattern, boh, boh_flags, eoh, eoh_flags, where,
                header_body_separator="¤", db_files_mode='r',
                record_positions_offset=0):
    """Search for matches of a TDBSF search expression among a set of files.

    Search for matches of 'expr' in the TDBSF database defined by
    'pattern', 'boh', 'boh_flags', 'eoh' and 'eoh_flags'.

    This is a convenience function that creates a Crawler instance
    and calls its 'search' method with the appropriate parameters. It
    is mainly provided for backward-compatibility.

    """
    crawler = Crawler(pattern, boh, boh_flags, eoh, eoh_flags,
                      db_files_mode=db_files_mode,
                      record_positions_offset=record_positions_offset)
    return crawler.search(expr, where, header_body_separator)


def test1(root_dir, quiet=False):
    pattern = os.path.join(root_dir, "tests", "*.tdb")
    boh, boh_flags = r"^={76}\^···\n", "MULTILINE"
    eoh, eoh_flags = r"^-{76}\^···\n", "MULTILINE"
    search_expr = r"""
    "[sS]haza+m !" & (bla | ("abc

azeaze" & "plonk !%IGNORECASE")) & "pou[eê]t%I" ~3 def ~4 abc ~5 plonk ~6
"pou+i+c !" | fin ~3 header ~2 file2
    """
    where = "header_or_body"

    # New interface
    crawler = Crawler(pattern, boh, boh_flags, eoh, eoh_flags)
    res = crawler.search(search_expr, where)
    # The same search with the old interface
    res2 = lookup_expr(search_expr, pattern, boh, boh_flags, eoh, eoh_flags,
                       where)

    assert res == res2 == \
        (1, [[os.path.join(root_dir, 'tests', 'file2.tdb'),
              [((23, 104), (138, 219)), ((288, 369), (387, 468)),
               ((533, 614), (638, 719)), ((831, 912),
        (934, 1015)), ((1099, 1180), (1244, 1325))], [((1099, 1180,
        1244, 1325, 1535), ('header', ['~', [(36, 39), (43, 49), (55,
        60)]]))]]]), (res, res2)

    if not quiet:
        print(res)


def test(root_dir, quiet=False):
    """Quick regression test of this module.

    The test requires database files located at
    <root_dir>/tests/*.tdb. This can be satisfied if 'root_dir' is a
    possibly empty, absolute or relative path to the root directory
    of the TDBSF distribution.

    From the shell command line, assuming the current directory is
    the root directory of the TDBSF distribution, the test can be run
    with:

      python3 -m tdbsf.tdbsf_api

    """
    retcode = 1
    try:
        test1(root_dir, quiet=quiet)
    except TDBSFGenericException as exc_instance:
        print(exc_instance.complete_message(), file=sys.stderr)
    except (os.error, IOError) as e:
        exc_msg = e.strerror
        if hasattr(e, "filename"):
            exc_msg = "{msg}: {file}".format(msg=exc_msg, file=e.filename)
        print("OS error: {0}".format(exc_msg), file=sys.stderr)
    except Exception:
        traceback.print_exc(file=sys.stderr)
        print('\n' + tw.fill(dedent(
                """\
            TDBSF unhandled exception! Please report precisely how you
            triggered it to Florent, with the traceback printed above this
            message!""")), file=sys.stderr)
    else:
        retcode = 0

    return retcode


if __name__ == "__main__":
    try:
        root_dir = sys.argv[1]
    except IndexError:
        root_dir = ""

    retcode = test(root_dir)
    assert isinstance(retcode, int), retcode
    sys.exit(retcode)
