# -*- coding: utf-8 -*-

# tdbsf_front_end.py --- Outer interface to the TDBSF "search engine"
# Copyright (c) 2001, 2004, 2009, 2013 Florent Rougon
#
# This file is part of TDBSF.
#
# TDBSF is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2, or (at your option)
# any later version.
#
# TDBSF is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public
# License for more details.
#
# You should have received a copy of the GNU General Public License
# along with TDBSF; see the file COPYING. If not, write to the Free
# Software Foundation, Inc., 59 Temple Place - Suite 330,
# Boston, MA 02111-1307, USA.

"""Front-ends to the Trivial Database Search Facility.

Currently, this module holds a GNU Emacs interface and a dumb (for
demonstrating purposes) text interface.

"""


import sys, os, locale, getopt, traceback, textwrap
from textwrap import dedent

from . import tdbsf_api
from . import error as TDBSFGenericException

progname = os.path.basename(sys.argv[0])
from . import version as progversion
version_blurb = """Written by Florent Rougon.

Copyright (C) 2001-2013  Florent Rougon
This is free software; see the source for copying conditions.  There is NO
warranty; not even for MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE."""

usage = """Usage: {progname} [OPTION]... EXPR PATTERN BOH EOH
Perform a TDBSF query.

EXPR                        search expression for the query
PATTERN                     shell-style globbing pattern to determine the
                            set of files to examine
BOH                         Python regular expression specifying the
                            beginning of header of a record
EOH                         Python regular expression specifying the end
                            of header of a record (after which starts the
                            the body of the record)

Options:
      --boh-flags=FLAGS     flags to use for the regular expression search
                            performed when looking for the Beginning Of Header
      --eoh-flags=FLAGS     flags to use for the regular expression search
                            performed when looking for the End Of Header
      --crlf-to-lf-in-args  convert CRLF to LF in command-line arguments
                            before passing them to the engine
  -m, --db-files-mode=MODE  mode to use when opening database files; must be
                            a valid mode string for Python's open() function,
                            for instance "r". Don't use this option unless
                            you really know what you are doing.
  -e, --emacs-friendly-output
                            generate output in a format that is easy to parse
                            in Emacs Lisp
      --header-body-separator=SEPARATOR
                            separator string to use between the header and the
                            body in the string where the search expression is
                            evaluated when the value given to --search-string
                            is header_plus_body
  -s, --search-string=WHERE part of each record where the search expression
                            is to be evaluated; permitted values are header,
                            body, header_plus_body and header_or_body.
      --help                display this message and exit
      --version             output version information and exit

Return:
   - 0 in case of normal termination;
   - a non-zero exit status if an error is encountered.""".format(
    progname=progname)

tw = textwrap.TextWrapper(width=80, break_long_words=False,
                          break_on_hyphens=False)


def build_translate_func(where):
    # Return a function converting match spans (i.e. (start, end) tuples) to
    # absolute positions in the file.
    if where == "header":
        def translate_func(match_span, record_info, lhbs):
            shift =  record_info[1]
            return (match_span[0] + shift, match_span[1] + shift)
    elif where == "body":
        def translate_func(match_span, record_info, lhbs):
            shift =  record_info[3]
            return (match_span[0] + shift, match_span[1] + shift)
    else:                               # 'where' must be "header_plus_body"
        def translate_func(match_span, record_info, lhbs):
            def translate_start_offset(offset, record_info, lhbs):
                abs_offset = offset + record_info[1]
                if abs_offset <= record_info[2]:
                    return abs_offset
                elif abs_offset < record_info[2] + lhbs:
                    return record_info[2]
                else:
                    return abs_offset + \
                           (record_info[3] - record_info[2]) - lhbs
            def translate_end_offset(offset, record_info, lhbs):
                abs_offset = offset + record_info[1]
                if abs_offset <= record_info[2]:
                    return abs_offset
                elif abs_offset <= record_info[2] + lhbs:
                    return record_info[3]
                else:
                    return abs_offset + \
                           (record_info[3] - record_info[2]) - lhbs

            return (
                translate_start_offset(match_span[0], record_info, lhbs),
                translate_end_offset(match_span[1], record_info, lhbs))

    return translate_func


def convert_match_spans_to_absolute(record_info, match_spans, lhbs,
                                    translate_func):
    if isinstance(match_spans, list):
        for i in range(len(match_spans[1])):
            match_spans[1][i] = convert_match_spans_to_absolute(
                record_info, match_spans[1][i], lhbs, translate_func)
        return match_spans
    elif isinstance(match_spans, tuple):
        return translate_func(match_spans, record_info, lhbs)
    else:
        return match_spans              # Should be a single 1


def flatten_match_spans_recurs(match_spans, res):
    if isinstance(match_spans, list):
        if match_spans[0] == '~':
            res.append(match_spans)
        else:                           # match_spans[0] must be '&'
           for i in range(len(match_spans[1])):
               index = 0
               elt = match_spans[1][index]
               if isinstance(elt, list):
                   flatten_match_spans_recurs(elt, res)
                   del match_spans[1][index]
               elif elt == 1:
                   del match_spans[1][index]
               else:
                   index += 1
           if match_spans[1]:           # If there are remaining elements
               res.append(match_spans)
    elif isinstance(match_spans, tuple):
        res.append(['&', [match_spans]]) # "fake &" (one operand only)
    # match_spans could also be a single 1, in case we add nothing to 'res'.


def flatten_match_spans(abs_match_spans):
    # Returns a list containing only ['&', [tuple, tuple, tuple...]] and
    # ['~', [tuple, tuple, tuple...]] elements.
    res = []
    flatten_match_spans_recurs(abs_match_spans, res)
    return res


def emacs_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                    search_string, header_body_separator, db_files_mode,
                    record_positions_offset):
    output = sys.stdout
    error = None

    try:
        crawler = tdbsf_api.Crawler(
            pattern, boh, boh_flags, eoh, eoh_flags, db_files_mode=db_files_mode,
            record_positions_offset=record_positions_offset)
        search_res = crawler.search(expr, search_string, header_body_separator)
    except TDBSFGenericException as exc_instance:
        error = exc_instance.complete_message()
    except (os.error, IOError) as e:
        exc_msg = e.strerror
        if hasattr(e, "filename"):
            exc_msg = "{msg}: {file}".format(msg=exc_msg, file=e.filename)
        error = "OS error: {0}".format(exc_msg)
    except Exception:
        error = traceback.format_exc() + '\n' + tw.fill(dedent(
                """\
            TDBSF unhandled exception! Please report precisely how you
            triggered it to Florent, with the traceback printed above this
            message!"""))

    if error is not None:
        output.write("TDBSF error:\n\n" + error)
    else:
        output.write("({0} . [".format(search_res[0])) # number of matches
        for file_res in search_res[1]:
            # Absolute pathname for the current file
            #   -> quote all backslashes for ELisp; it is better than
            #      translating them into slashes, I think, because it still
            #      allows backslashes in Unix file paths.
            output.write('("{0}" . [['.format(
                    file_res[0].replace("\\", "\\\\")))

            # First, output information about all records contained in the
            # file (not only the records that matched the search expression):
            # boh start, boh end, eoh start and eoh end.
            all_records_info = []
            for record in file_res[1]:
                # We add 1 to every offset in the file contents because buffer
                # positions start at 1 in Emacs.
                all_records_info.append(
                "[({0} . {1}) ({2} . {3})]".format(record[0][0] + 1,
                                                   record[0][1] + 1,
                                                   record[1][0] + 1,
                                                   record[1][1] + 1))
            # Write and terminate the array, then start the array containing
            # information about matching records.
            output.write(' '.join(all_records_info) + "]\n[")

            info_about_matches = []
            for match in file_res[2]:
                # Convert match spans to absolute offsets in the file and make
                # them Emacs-friendly (meaning the first byte in the file is
                # at offset 1).
                translate_func = build_translate_func(match[1][0])
                abs_match_spans = convert_match_spans_to_absolute(
                    match[0], match[1][1], len(header_body_separator),
                    translate_func)
                # Flatten (and thus degrade, alas) fontification information
                flat_match_spans = flatten_match_spans(abs_match_spans)

                # Turn fontification information into an ELisp structure
                match_info = []
                for group in flat_match_spans:
                    match_spans_list = []
                    for match_span in group[1]:
                        match_spans_list.append(
                "({0} . {1})".format(*map(str, match_span)))

                    match_info.append(
                        # (OPERATOR . MATCH-SPANS)
                        # where OPERATOR can be "~" or "&"
                        '("{0}" . [{1}])'.format(group[0],
                                                 ' '.join(match_spans_list)))

                # Turn record-match information into an ELisp structure
                info_about_matches.append("([{0}] . [{1}])".format(
                        # Record information
                        ' '.join(map(str, match[0])),
                        # Fontification (match) information
                        ' '.join(match_info)))

            output.write(' '.join(info_about_matches) + "]])\n")
        output.write("])")

    # Useful for debugging...
#     print ("\nexpr: %s\npattern: %s\nboh: %s\neoh: %s\nsearch_string: %s\n"
#            + "header_body_separator: %s") % (expr, pattern, boh, eoh,
#                                              search_string,
#                                              header_body_separator)
#     print "len(boh): %d\nlen(eoh): %s" % (len(boh), len(eoh))

    output.write("\n\n*** End of TDBSF Emacs-oriented output ***\n")


def dumb_text_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                        search_string, header_body_separator, db_files_mode,
                        record_positions_offset):
    # We do not handle exceptions in this dumb interface to keep it as simple
    # as possible.
    crawler = tdbsf_api.Crawler(
        pattern, boh, boh_flags, eoh, eoh_flags, db_files_mode=db_files_mode,
        record_positions_offset=record_positions_offset)
    search_res = crawler.search(expr, search_string, header_body_separator)

    for file_res in search_res[1]:
        # file_res[0] gives the file path
        print("{0} -> {1} record(s) matched:\n".format(file_res[0],
                                                       len(file_res[2])))
        for match in file_res[2]:
            print("  Record information: {0}".format(match[0]))
            print("  Fontification information: {0}\n".format(match[1]))
    print("\nTotal: {0} matches".format(search_res[0]))


def main():
    locale.setlocale(locale.LC_ALL, '')

    try:
        opts, args = getopt.getopt(sys.argv[1:], "em:s:",
                                   ["boh-flags=",
                                    "crlf-to-lf-in-args",
                                    "db-files-mode=",
                                    "emacs-friendly-output",
                                    "eoh-flags=",
                                    "header-body-separator=",
                                    "search-string=",
                                    "help",
                                    "version"])
    except getopt.GetoptError as message:
        sys.stderr.write(usage + '\n')
        sys.exit(2)

    # Let's start with the options that don't require any non-option argument
    # to be present
    for option, value in opts:
        if option == "--help":
            print(usage)
            sys.exit(0)
        elif option == "--version":
            print("{name} {version}\n{blurb}".format(
                    name=progname, version=progversion, blurb=version_blurb))
            sys.exit(0)

    # Now, require a correct invocation.
    if len(args) != 4:
        sys.stderr.write(usage + '\n')
        sys.exit(2)

    # These are already (Unicode) strings in Python 3, no need to decode them
    expr, pattern, boh, eoh = args

    # Default values for options
    boh_flags = ""
    eoh_flags = ""
    # "Universal newline" mode is enabled by default for text mode in Python 3
    db_files_mode = 'r'
    emacs_friendly_output = False
    search_string = "header_or_body"
    header_body_separator = '¤'         # CURRENCY SIGN
    crlf_to_lf_in_args = False

    for option, value in opts:
        if option == "--boh-flags":
            boh_flags = value
        elif option == "--crlf-to-lf-in-args":
            crlf_to_lf_in_args = True
        elif option == "--eoh-flags":
            eoh_flags = value
        elif option in ("-m", "--db-files-mode"):
            if "w" in value:
                sys.exit("TDBSF error:\n\nRefusing to start because the value "
                         "given for option -m (--db-files-mode) contains 'w', "
                         "which could erase the database files.")
            db_files_mode = value
        elif option in ("-e", "--emacs-friendly-output"):
            emacs_friendly_output = True
        elif option in ("-s", "--search-string"):
            search_string = value
        elif option == "--header-body-separator":
            header_body_separator = value
        else:
            assert False, \
                "{prog}: unexpected option received from the getopt module: " \
                "'{option}' (probably a bug in {prog})".format(
                prog=progname, option=option)

    if crlf_to_lf_in_args:
        expr = expr.replace("\r\n", "\n")
        boh = boh.replace("\r\n", "\n")
        eoh = eoh.replace("\r\n", "\n")
        header_body_separator = header_body_separator.replace("\r\n", "\n")

    if emacs_friendly_output:
        # Within Emacs, buffer positions start at 1, not 0, hence the last
        # argument's value to offset all positions specifying the beginning of
        # the "boh" part of a record, the end of this part, the beginning of
        # its "eoh" part, its end, and the end of the record's body.
        emacs_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                        search_string, header_body_separator, db_files_mode,
                        1)
    else:
        # In this dumb interface, we let the file positions start at 0.
        dumb_text_interface(expr, pattern, boh, boh_flags, eoh, eoh_flags,
                            search_string, header_body_separator,
                            db_files_mode, 0)

    sys.exit(0)
